@@ -6,16 +6,17 @@ Functions used to render and return discovery responses to Envoy proxies.
 
 The templates are configurable. `todo See ref:Configuration#Templates`
 """
-import zlib
 import yaml
 from yaml.parser import ParserError, ScannerError
 from enum import Enum
+from typing import Union
+from collections import namedtuple
 from starlette.exceptions import HTTPException
 from sovereign import XDS_TEMPLATES, config
+from sovereign.utils.version_info import compute_hash
 from sovereign.logs import LOG
-from sovereign.statistics import stats
 from sovereign.context import safe_context
-from sovereign.schemas import XdsTemplate, DiscoveryRequest
+from sovereign.schemas import XdsTemplate, DiscoveryRequest, ProcessedTemplate
 
 try:
     import sentry_sdk
@@ -33,31 +34,28 @@ except KeyError:
 # Create an enum that bases all the available discovery types off what has been configured
 discovery_types = (_type for _type in sorted(XDS_TEMPLATES['__any__'].keys()))
 # noinspection PyArgumentList
-DiscoveryTypes = Enum('DiscoveryTypes', {t: t for t in discovery_types})
+all_types = {t: t for t in discovery_types}
+DiscoveryTypes = Enum('DiscoveryTypes', all_types)
+
+NotModified = namedtuple('NotModified', ['version_info', 'resources'])
 
 
-@stats.timed('discovery.version_hash_ms')
-def version_hash(*args) -> str:
-    """
-    Creates a 'version hash' to be used in envoy Discovery Responses.
-    """
-    data: bytes = repr(args).encode()
-    version_info = zlib.adler32(data) & 0xffffffff  # same numeric value across all py versions & platforms
-    return str(version_info)
+def select_template(request: DiscoveryRequest, discovery_type: DiscoveryTypes, templates=None) -> XdsTemplate:
+    if templates is None:
+        templates = XDS_TEMPLATES
+    version = request.envoy_version
+    selection = 'default'
+    for v in templates.keys():
+        if version.startswith(v):
+            selection = v
+    selected_version = templates[selection]
+    try:
+        return selected_version[discovery_type]
+    except KeyError:
+        raise KeyError(f'Unable to get {discovery_type} for template version "{selection}". Envoy client version: {version}')
 
 
-def load_template(request, type_) -> XdsTemplate:
-    return config.select_template(request.envoy_version)[type_]
-
-
-async def process_template(template, context):
-    if template.is_python_source:
-        return {'resources': list(template.code.call(**context))}
-    else:
-        return await template.content.render_async(**context)
-
-
-async def response(request: DiscoveryRequest, xds_type: DiscoveryTypes, host: str = 'none'):
+async def response(request: DiscoveryRequest, xds_type: DiscoveryTypes) -> Union[ProcessedTemplate, NotModified]:
     """
     A Discovery **Request** typically looks something like:
 
@@ -88,40 +86,40 @@ async def response(request: DiscoveryRequest, xds_type: DiscoveryTypes, host: st
 
     :param request: An envoy Discovery Request
     :param xds_type: what type of XDS template to use when rendering
-    :param host: the host header that was received from the envoy client
     :return: An envoy Discovery Response
     """
-    template: XdsTemplate = load_template(request, xds_type)
+    template: XdsTemplate = select_template(request, xds_type)
     context: dict = safe_context(request, template)
 
-    config_version = '0'
+    config_version = None
     if config.cache_strategy == 'context':
-        config_version = version_hash(context, template.checksum, request.node.common, request.resources)
-        if config_version == request.version_info:
-            return {'version_info': config_version}
-
-    content = await process_template(
-        template=template,
-        context=dict(
-            discovery_request=request,
-            host_header=host,
-            resource_names=request.resources,
-            **context
+        config_version = compute_hash(
+            context,
+            template.checksum,
+            request.node.common,
+            request.resource_names,
+            request.desired_controlplane
         )
-    )
-
-    # We can determine the version number before deserializing the YAML string into python
-    if config.cache_strategy == 'content':
-        config_version = version_hash(content)
         if config_version == request.version_info:
-            return {'version_info': config_version}
+            return NotModified(version_info=config_version, resources=[])
 
-    # This is the most expensive operation, I think, so it's performed as late as possible.
+    context = dict(
+        discovery_request=request,
+        host_header=request.desired_controlplane,
+        resource_names=request.resources,
+        **context
+    )
+    content = await template(**context)
     if not template.is_python_source:
         content = deserialize_config(content)
 
-    content['version_info'] = config_version
-    return filter_resources(content, request.resources)
+    if config.cache_strategy == 'content':
+        config_version = compute_hash(content)
+        if config_version == request.version_info:
+            return NotModified(version_info=config_version, resources=[])
+
+    resources = filter_resources(content['resources'], request.resources)
+    return ProcessedTemplate(resources=resources, type_url=request.type_url, version_info=config_version)
 
 
 def deserialize_config(content):
@@ -150,20 +148,17 @@ def deserialize_config(content):
     return envoy_configuration
 
 
-def filter_resources(conf, requested):
+def filter_resources(generated, requested):
     """
     If Envoy specifically requested a resource, this removes everything
     that does not match the name of the resource.
     If Envoy did not specifically request anything, every resource is retained.
     """
-    ret = dict()
-    ret['version_info'] = conf['version_info']
-    ret['resources'] = [
+    return [
         resource
-        for resource in conf.get('resources', [])
+        for resource in generated
         if resource_name(resource) in requested
     ]
-    return ret
 
 
 def resource_name(resource):
